#include "CTask.h"

//-------------------------------------------
// Ctor 
//-------------------------------------------
CTask::CTask():
    _parent(0),
    _childCounter(0),
    _id(0),
    _hasInitThreadAttrs(false),
    _isRootTask(false),
    _taskQueue(0)
{

    if(false == _hasInitThreadAttrs)
        initThreadAttributes();
}

//-------------------------------------------
// Copy Ctor
//-------------------------------------------
CTask::CTask(const CTask & in) :
    _parent(in._parent),
    _childCounter(in._childCounter),
    _id(in._id),
    _hasInitThreadAttrs(in._hasInitThreadAttrs),
    _isRootTask(in._isRootTask),
    _taskQueue(in._taskQueue)
{

    if(false == _hasInitThreadAttrs)
        initThreadAttributes();
}

//-------------------------------------------
// Dtor
//-------------------------------------------
CTask::~CTask()
{
    //Lock mutex
    pthread_mutex_lock(&_child_decr_mutex);
    
    // Delete children
    for (int i = 0; i < _children.size(); i++)
    {
	if (0 != _children[i])
	{
	    delete _children[i];
	    _children[i] = 0; 
	}
    }
    _children.clear();
    
    // Unlock mutex
    pthread_mutex_unlock(&_child_decr_mutex);

    // Delete mutex
    if(true == _hasInitThreadAttrs)
    {
        pthread_mutex_destroy(&_child_decr_mutex);
    }
}

//-------------------------------------------
// Equal operator
//-------------------------------------------
CTask & CTask::operator=(const CTask & in)
{
    if(this == &in) return *this;
    _parent = in._parent;
    _childCounter = in._childCounter;
    _id = in._id;
    _hasInitThreadAttrs = in._hasInitThreadAttrs;
    
    if(false == _hasInitThreadAttrs)
        initThreadAttributes();

    return *this;
}

//-------------------------------------------
// Initialize threading attributes
//-------------------------------------------
void CTask::initThreadAttributes()
{
    pthread_mutex_init(&_child_decr_mutex, NULL);
    _hasInitThreadAttrs = true;
}

//-------------------------------------------
// Decrement child count and return result
//-------------------------------------------
int CTask::decrChildCounter() 
{ 
    int retVal = 0;

    //Lock update mutex
    pthread_mutex_lock(&_child_decr_mutex);
    
    //Save decremented counter in local variable
    retVal = --_childCounter;

    //Unlock update mutex
    pthread_mutex_unlock(&_child_decr_mutex);

    return retVal; 
}

//-------------------------------------------
// Virtual function
//-------------------------------------------
void CTask::Work()
{
    return;    
}

