#ifndef CTASK_H
#define CTASK_H
#include <vector>
#include <pthread.h>

class TaskQueue;

class CTask
{
private:
    CTask* _parent;
    std::vector<CTask*> _children; 
    volatile int _childCounter;
    int _id;
    bool _hasInitThreadAttrs;
    bool _isRootTask;
    TaskQueue* _taskQueue; 
    pthread_mutex_t _child_decr_mutex;
public:
    //Ctor
    CTask();
    CTask(const CTask & in);
    virtual ~CTask();
    CTask & operator=(const CTask & in);

    //Work
    void initThreadAttributes();
    int decrChildCounter(); 
    virtual void Work();

    //Getters/Setters
    CTask* getParent() const { return _parent; }
    int getChildCounter() const { return _childCounter; }
    std::vector<CTask*> & getChildren() { return _children; }
    int getId() const { return _id; }
    bool isRootTask() const { return _isRootTask; }
    TaskQueue* getTaskQueue() const { return _taskQueue; }
    
    void setIsRootTask(const bool in) { _isRootTask = in; }
    void setParent(CTask* const in) { _parent = in; }
    void setChildCount(const int in) { _childCounter = in; }
    void incrementChildCount() { _childCounter++; }
    void setId(const int id) { _id = id; }
    void addChild(CTask* task) { _children.push_back(task); }
    void setTaskQueue(TaskQueue* queue) { _taskQueue = queue; }
};
#endif
