#ifndef CTHREAD_H
#define CTHREAD_H
#ifdef _MFC_VER
#pragma once
#endif
#include <string>
#include <vector>
#include <pthread.h>

class CTask;
class ThreadHandler;

class CThread
{
    private:
        pthread_t _thread;
        pthread_attr_t _thread_attribute;
        pthread_mutex_t _mutex_suspend;
        pthread_mutex_t _mutex_state;
        pthread_cond_t _cond_resume;
        bool _hasInitThreadAttribute;
        volatile bool _suspend;
        volatile bool _exit;
        CTask* _task;
        int _id;
        ThreadHandler* _threadHandler;
    public:
        //Ctor
        CThread();
        CThread(const CThread & in);
        CThread & operator=(const CThread & in);
        virtual ~CThread();
        //End ctor

        //Work
        void initThreadAttribute();
        void create();
        bool resume(void (*ThreadFunction)(void*), void* dataPtr);
        bool resume(CTask* task);
        bool suspend();
        bool join();
        void print() const;
        static void* theWork(void* dataPtr);
        //End Work

        //Getters/Stters
        pthread_t getThread() const { return _thread; }
        pthread_mutex_t & getMutex() { return _mutex_suspend; }
        pthread_cond_t & getCond() { return _cond_resume; }
        CTask* getTask() const { return _task; }        
        int getId() const { return _id; }
        ThreadHandler* getThreadHandler() const { return _threadHandler; }

        bool exit() const volatile { return _exit; }
        bool isSuspended();
        
        void setTask(CTask* in);
        void setExit(bool in) volatile { _exit = in; }
        void setId(int id) { _id = id; } 
        void setThreadHandler(ThreadHandler* in) { _threadHandler = in; }
        //End Getters/Setters
};
#endif

