#ifndef TASK_QUEUE_H
#define TASK_QUEUE_H
#include <queue>
#include <pthread.h>

class CTask;

class TaskQueue {
private:
  //    std::queue< CTask* > _queue;
    std::vector< CTask* > _rootTasks;
    volatile int _numRootTasks;
    volatile int _position; 
    std::vector< CTask* > _queue; 
    pthread_mutex_t _mutex_root_tasks;
    pthread_mutex_t _mutex;
    pthread_mutex_t _mutex_main_thread_barrier; 
    pthread_cond_t _cond_main_thread_barrier; 
    volatile bool _doneDistributingWork;
    bool _hasInitThreadAttrs;        
public:    
    //Ctor
    TaskQueue();
    TaskQueue(const TaskQueue & in);
    TaskQueue & operator=(const TaskQueue & in); 
    virtual ~TaskQueue();

    //Thread attributes
    void initThreadAttributes();
    pthread_mutex_t & getBarrierMutex() { return _mutex_main_thread_barrier; }

    //Work
    void addRootTask(CTask* in);
    std::vector< CTask* > & getRootTasks() { return _rootTasks; }
    void suspendMainThread();
    int signalMainThread();
    void resetQueue();
//    void setRootTask(CTask* task);
    void incrementRootCounter() { _numRootTasks++; }
    int decrementRootCounter();
    int getRootCounter() const { return _numRootTasks; }
    bool hasRootTasks() const { return (_rootTasks.size() > 0); }
    void push(CTask* task);
    CTask* pop();
    bool empty();
    void traverseQueue();

    std::vector< CTask* > & getQueue() { return _queue; }
};
#endif
