#ifndef THREAD_HANDLER_H
#define THREAD_HANDLER_H
#ifdef _MFC_VER
#pragma once
#endif

#include <string>
#include <vector>
#include <stack>
#include <pthread.h>

// Global variable for number of threads
extern int g_numThreads;

class TaskQueue;
class CThread;
class CTask;

class ThreadHandler
{
private:
    static ThreadHandler* _instance;
    pthread_mutex_t _mutex;
    pthread_mutex_t _mutex_task_queues;
    pthread_mutex_t _mutex_main_thread_barrier;
    pthread_cond_t _cond_main_thread_barrier;
    volatile bool _doneDistributingWork;
    std::stack< TaskQueue* > _taskQueues;
    std::vector< CThread* > _threadPool;
    std::stack< CThread* > _readyThreadPool;
    bool _hasInitAttributes;
    int _timerCounter;
public:
    //Ctor
    static ThreadHandler* Instance();
    ThreadHandler();
    ThreadHandler(const ThreadHandler & in);
    ThreadHandler & operator=(const ThreadHandler & in);
    virtual ~ThreadHandler();
    //End ctor

    //Work
    void addTaskQueue(TaskQueue* in);
    void addThreads(const int num);
    void initAttributes();
    void distributeWork(TaskQueue* queue);
    void joinThreads();
    void pushReadyThread(CThread* thread);
    void killThreads(int numToKill);
    CThread* popReadyThread();
    bool readyThreadPoolEmpty();
    //End Work

    //Getters/Stters
    std::stack< TaskQueue* > & getTaskQueues() { return _taskQueues; }
    std::vector< CThread* > & getThreadPool() { return _threadPool; }
    std::stack< CThread* > & getReadyThreadPool() { return _readyThreadPool; }
    TaskQueue * getQueue();
    CTask * getTask();
    int getNumThreads() const { return _threadPool.size(); }
    //End Getters/Setters

    //Timing and dynamic thread handling
    void addTiming(int & index, const double time);
};

#endif

