#include "ThreadTimer.h"

ThreadTimer* ThreadTimer::_instance = 0;

//--------------------------------------------------------------
// Create a single instance of the timer
//--------------------------------------------------------------
ThreadTimer* ThreadTimer::Instance()
{
    if (0 == _instance)
    {
	_instance = new ThreadTimer();
    }
    return _instance; 
}

//--------------------------------------------------------------
// Ctor
//--------------------------------------------------------------
ThreadTimer::ThreadTimer() : 
    _start( std::clock() ), 
    _difference( 0 ),
    _numTimings( 3 ),
    _timingCounter( 0 ),
    _optNumThreads( 1 ),
    _optNumSet( false )
{}

//--------------------------------------------------------------
// Copy Ctor
//--------------------------------------------------------------
ThreadTimer::ThreadTimer(const ThreadTimer & in) :
    _start( in._start ), 
    _difference( in._difference ),
    _numTimings( in._numTimings ),
    _timingCounter( in._timingCounter ),
    _optNumThreads( in._optNumThreads ),
    _optNumSet( in._optNumSet ),
    _timings( in._timings ),
    _timers( in._timers)
{

}

//--------------------------------------------------------------
// Equal operator
//--------------------------------------------------------------
ThreadTimer & ThreadTimer::operator=(const ThreadTimer & in)
{
    if(this == &in) return *this;
    _start = in._start; 
    _difference = in._difference;
    _numTimings = in._numTimings;
    _timingCounter = in._timingCounter;
    _optNumThreads = in._optNumThreads;
    _optNumSet = in._optNumSet;
    _timings = in._timings;
    _timers = in._timers;
    return *this;   
}


//--------------------------------------------------------------
// Dtor
//--------------------------------------------------------------
ThreadTimer::~ThreadTimer()
{
    for (int i = 0; i < _timers.size(); i++)
    {
	if (0 !=_timers[i])
	{
	    delete _timers[i];
	}
    }    
}

//--------------------------------------------------------------
// Get thread count with fastest calculation time
//--------------------------------------------------------------
int ThreadTimer::get_opt_thread_num() const
{
    // Quick return when opt num threds is chosen
    if(true == _optNumSet)
	return _optNumThreads;

    // Return if no timings are done
    if(0 == _timings.size()) 
	return -1;

    // Get optimal number of threads 
    int nThreads = _timings[0].first; 
    double fastest = _timings[0].second;
    for (int i = 1; i < _timings.size(); i++)
    {
	double tmp = _timings[i].second;
	if (tmp < fastest)
	{
	    fastest = tmp;
	    nThreads = _timings[i].first; 
	}
    }    
    return nThreads;
}

//--------------------------------------------------------------
// Get the corresponding thread timer and return corrected index
//--------------------------------------------------------------
ThreadTimer* ThreadTimer::get_thread_timer(int & index)
{
    // Add a new timer and return iits index
    if (index >= _timers.size())
    {
	_timers.push_back(new ThreadTimer());
	index = _timers.size() - 1; 
    }
    return _timers[index];
}

//--------------------------------------------------------------
// Save timings for current number of threads
//--------------------------------------------------------------
int ThreadTimer::save_time(const int nThreads, const double time)
{
    bool found = false; 
    for (int i = 0; i < _timings.size(); i++)
    {
	double tmpThreads = _timings[i].first;
	if (tmpThreads == nThreads)
	{
	    // Calculate average time
	    double tmpTime = _timings[i].second + time;
	    _timings[i].second = tmpTime;
	    _timingCounter++;

//	    // Add threads when done timing for current number of threads
//	    if (_timerCounter >= _numTimings)
//	    {
//		_timerCounter = 0;
//		ThreadHandler::Instance()->DynamicAddThreads();
//	    }
//
	    found == true; 
	    break;
	}
    }    

    // Add timings for current num of threads if not found 
    if (false == found)
    {
	_timings.push_back(std::make_pair(nThreads, time));
    }
    return -1;
}
