#ifndef THREAD_TIMER_H
#define THREAD_TIMER_H
#ifdef _MFC_VER
#pragma once
#endif

#include <sys/time.h>
#include <ctime>
#include <utility>
#include <vector>

class ThreadTimer
{
private:
    static ThreadTimer* _instance;
    std::clock_t _start;
    timeval _time_start;
    double _difference; 
    int _numTimings;
    int _timingCounter;
    int _optNumThreads;
    bool _optNumSet;
    std::vector< std::pair< int, double > > _timings; 
    std::vector< ThreadTimer* > _timers;

public:
    //Ctor
    static ThreadTimer* Instance();
    ThreadTimer();
    ThreadTimer(const ThreadTimer & in);
    ThreadTimer & operator=(const ThreadTimer & in);
    virtual ~ThreadTimer();
    // End ctor

    void start_ms_timer() 
    { 
	_start = std::clock(); 
	_difference = 0; 
    }

    // Returns time in miliseconds
    double stop_ms_timer()	
    { 
	_difference = std::difftime(std::clock(), _start); 
	return _difference; 
    }

    void start_us_timer()
    {
	gettimeofday(&_time_start,0);
    }
    
    // Returns time in microseconds
    unsigned long  stop_us_timer()
    {
	timeval end; 
	gettimeofday(&end, 0);
	return ((unsigned long)(end.tv_sec) * 1000000 + (unsigned long)(end.tv_usec)) - 
	    ((unsigned long)(_time_start.tv_sec) * 1000000 + (unsigned long)(_time_start.tv_usec));
       
    }

    double last_diff() const { return _difference; }
    void set_num_timings(const int num) { _numTimings = num; }
    int get_num_timings() const { return _numTimings; }
    void set_opt_thread_num(const int num) { _optNumThreads = num; _optNumSet = true; }

    int get_opt_thread_num() const;
    ThreadTimer* get_thread_timer(int & index);
    int save_time(const int nThreads, const double time); 
};

#endif
