#include <stdio.h>
#include <vector>
#include <cstdlib>
#include <unistd.h>
#include <time.h>
#include <iostream>
#include <fstream>

#include <boost/tokenizer.hpp>

#include "CTask.h"
#include "CThread.h"
#include "TaskQueue.h"
#include "ThreadHandler.h"
#include "ThreadTimer.h"

using namespace std;

int g_localNumThreads = 0;
int g_numTasks = 0;
int g_subQueues = 10;
int g_counter = 0;

CTask* buildTaskTree(TaskQueue & taskQueue, CTask* parent_task, int depth);

//------------------------------------------------------------------
// Sleeps for a specified duration in microseconds
//------------------------------------------------------------------
int my_usleep(unsigned long us)
{
    // Convert to nanoseconds
    timespec time;
    time_t secs = (int) (us / 1000000L);
    us = us - (secs * 1000000);
    time.tv_sec = secs;
    time.tv_nsec = us * 1000L;

    // - return value of -1 means error or signal received
    // - 2nd parameter returns remaining sleep time
    // - Spin while there is more sleep time left
    while (nanosleep(&time, &time) == -1)
    {
        continue;
    }
    return 1;
}

//------------------------------------------------------------------
// A Test task
//------------------------------------------------------------------
struct TestTask : public CTask
{
    int var;
    bool isParent;
    bool createQueue;
    void setIfParent(bool in)
    {
        isParent = in;
    }

    //Override work
    void Work()
    {
//      if (true == createQueue && g_subQueues > 0)
        if(false)
        {
            g_subQueues--;

            g_counter = 100;

            TaskQueue taskQueue;
            buildTaskTree(taskQueue, 0, 0);
//          printf("!!!! Sub task distributing queue!!!\n");
            ThreadHandler::Instance()->distributeWork(&taskQueue);
//          printf("!!!! Sub task DONE distributing queue!!!\n");
            return;
        }

        int id = getId();
//        printf("--- TestTask %d, is parent: %d: Do unuseful work ---\n", id, (isParent)?1:0);
        var = 10;
        for(size_t i = 0; i < 10000; i++)
        {
            int tmp = var+10*50*i;
            int tmp2 = tmp + var++;
//            printf("Var: %d ", var++);
            var = tmp2 + var;
        }
//        printf("\n");
        if(var == 42)
            printf("VAR IS 42!!!!\n");
    }
};

//------------------------------------------------------------------
// Task that sleeps for a given amount of time
//------------------------------------------------------------------
struct SleepTask : public CTask
{
    string name;
    bool isWell;
    string parent;
    int nSeg;
    unsigned long sleep_usec;

    //Override work
    void Work()
    {
        long rest = (long)sleep_usec;
        long sum = 0L;
        timeval start, end;
        unsigned long tmp;
        gettimeofday(&start,0);
        do
        {
            for(unsigned long i = 0; i < rest; ++i)
                sum += 1;

            gettimeofday(&end,0);

            tmp = ((unsigned long)(end.tv_sec) * 1000000 + (unsigned long)(end.tv_usec)) -
                ((unsigned long)(start.tv_sec) * 1000000 + (unsigned long)(start.tv_usec));

            rest = sleep_usec - tmp;
        }while(sleep_usec > tmp);

//      printf("\n");


//        printf("ID: %d: %s slept for %ld microseconds. Sum: %ld, Rest: %ld\n", getId(), name.c_str(), sleep_usec, sum, rest);
//        my_usleep(sleep_usec);
//      usleep(sleep_usec);
    }
};




//------------------------------------------------------------------
// Recursively builds a test case tree
//------------------------------------------------------------------
CTask* buildTaskTree(TaskQueue & taskQueue, CTask* parent_task, int depth)
{
    if(g_counter <= (1 << depth))
        return 0;

    TestTask* task = new TestTask;
    task->setId(g_counter--);
    task->setParent(parent_task);
    task->createQueue = false;

    if (0 == g_counter%1000)
    {
//      printf("Task shall make subqueue!\n");
        task->createQueue = true;
    }

    if (0 == parent_task)
    {
        taskQueue.addRootTask(task);
        task->setIsRootTask(true);
    }
    else
    {
        parent_task->addChild(task);
    }

    bool isLeafNode = true;
    for(size_t i = 0; i < 3; i++)
    {
        if(0 != buildTaskTree(taskQueue, task, depth+1))
        {
            task->setChildCount(task->getChildCounter()+1);
            isLeafNode = false;
        }
    }
    if(true == isLeafNode)
    {
        taskQueue.push(task);
    }

    return task;
}

//------------------------------------------------------------------
// Generates the task queue three from a list of tasks
//------------------------------------------------------------------
void makeSleepTaskTree(std::vector< SleepTask* > & tasks, TaskQueue & taskQueue)
{
    printf("Making tree\n");
    SleepTask* task1;
    SleepTask* task2;
    for (int i = 0; i < tasks.size(); i++)
    {
        task1 = tasks[i];
        if (true == task1->isWell)
        {
            cout << "Well " << task1->name << "\n";
            taskQueue.push(task1);
        }
        else
        {
            bool found_task = false;
            for (int j = 0; j < tasks.size(); j++)
            {
                task2 = tasks[j];
                if (0 == task1->name.compare(task2->parent))
                {
                    cout << "Child: " << task2->name << " Parent: " << task1->name << "\n";
                    task1->incrementChildCount();
                    task1->addChild(task2);
                    task2->setParent(task1);
                    found_task = true;
                }
            }
            if(!found_task)
                printf("BUG!!!: Did not find a task!\n");
        }
        if (0 == task1->parent.compare("0"))
        {
            printf("Task %s is root task\n", task1->name.c_str());
            taskQueue.addRootTask(task1);
        }
    }
}

//------------------------------------------------------------------
// Reads the task file and generates a list of task queues
//------------------------------------------------------------------
void readTasksFile(string filename, std::vector< TaskQueue* > & taskQueues)
{
    ifstream my_file(filename.c_str(), ios::in);
    string line;
    std::vector< SleepTask* > tasks;

    // Create a tokenizer with ; delimiter
    typedef boost::tokenizer< boost::char_separator<char> > tokenizer;
    boost::char_separator<char> sep(";");

    int g_counter = 0;

    // create tasks
    if (my_file.is_open())
    {
        while (!my_file.eof())
        {
            // Get a line from the file
            getline(my_file, line);

            // Tokenize line
            tokenizer tokens(line, sep);

            // Get all tokens
            std::vector< string > strings;
            for (tokenizer::iterator iter = tokens.begin(); iter != tokens.end(); ++iter )
            {
                strings.push_back(*iter);
            }

            // Create task tree when new iteration is eng_countered
            if (5 != strings.size())
            {
                if (tasks.size() > 0)
                {
                    // Create a new queue
                    TaskQueue* queue = new TaskQueue();
                    taskQueues.push_back(queue);

                    // Build queue tree
                    makeSleepTaskTree(tasks, *queue);

                    // Reset the tasks list for next iteration
                    tasks.clear();

                    cout << "\n";
                }
                continue;
            }

            // Extract data
            string name = strings[0];
            bool isWell = (1 == atoi((strings[1]).c_str())) ? true : false;
            string parent = strings[2];
            int nseg = atoi((strings[3]).c_str());
            unsigned long usec = atol((strings[4]).c_str());

            // Print data
            printf("Name: %s, isWell: %d, Parent: %s, nseg: %d, usec: %ld\n",  name.c_str(), isWell, parent.c_str(), nseg, usec);

            // Create tasks
            SleepTask* task = new SleepTask();
            task->setId(g_counter++);
            task->name = name;
            task->isWell = isWell;
            task->parent = parent;
            task->nSeg = nseg;
            task->sleep_usec = usec;
            tasks.push_back(task);
        }
        my_file.close();
    }
}

//------------------------------------------------------------------
// Calculates the hight of a network
//------------------------------------------------------------------
int findHeight( SleepTask* task )
{
    if( true == task->isWell)
    {
        return 1;
    }
    int heightFromHere = 0;
    for(int i = 0; i < task->getChildren().size(); i++)
    {
        int tmpHeight = 0;
        tmpHeight = findHeight((SleepTask*)task->getChildren()[i]);
        if(tmpHeight >= heightFromHere)
        {
            heightFromHere = tmpHeight;
        }
    }
    return heightFromHere + 1;
}

//------------------------------------------------------------------
// Calculates the number of pipes and segments in a level in the network
//------------------------------------------------------------------
unsigned int sumSegments(int level, std::pair<int,unsigned int>* levelInfo, SleepTask* task)
{
    levelInfo[level].first += 1;
    levelInfo[level].second += task->nSeg;

    unsigned int numSegments = 0;
    for(int i = 0; i < task->getChildren().size(); i++)
    {
        numSegments += sumSegments(level+1, levelInfo, (SleepTask*)task->getChildren()[i]);
    }

    return numSegments + task->nSeg;
}

//------------------------------------------------------------------
// Calculates the number of pipes and time in a level in the network
//------------------------------------------------------------------
double sumTime(int level, std::pair<int,double>* levelInfo, SleepTask* task)
{
    levelInfo[level].first += 1;
    levelInfo[level].second += (double)task->sleep_usec;

    double time = 0;
    for(int i = 0; i < task->getChildren().size(); i++)
    {
        time += sumTime(level+1, levelInfo, (SleepTask*)task->getChildren()[i]);
    }

    return time + task->sleep_usec;
}

//------------------------------------------------------------------
// Calculates the minimum amount of time needed to calculate
// a whole tree in parallel under perfect conditions
// (by finding max calc time for each child level)
//------------------------------------------------------------------
double findMinimumPossibleTime(SleepTask* task)
{
    double max = 0;
    for(int i = 0; i < task->getChildren().size(); i++)
    {
        double time = findMinimumPossibleTime((SleepTask*)task->getChildren()[i]);
        max = (time > max) ? time : max;
    }

    return max + task->sleep_usec;
}

//------------------------------------------------------------------
// Calculates the theoretical formula
//------------------------------------------------------------------
void printCalculationTimes( std::vector< TaskQueue* > & taskQueues )
{
    // Traverse all queues
    double sumTs = 0.0;
    double sumTp = 0.0;
    for(int i = 0; i < taskQueues.size(); i++)
    {
	// Make fake super root task if more than 1 root in taskQueue
        SleepTask* rootToUse = 0;
        if(taskQueues[i]->getRootTasks().size() > 1)
        {
            SleepTask* superRoot = new SleepTask;
            superRoot->name = "SuperRoot";
            superRoot->isWell = false;
            superRoot->parent = "0";
            superRoot->nSeg = 1;
            superRoot->sleep_usec = 0;
            for(int j = 0; j < taskQueues[i]->getRootTasks().size(); j++)
            {
                superRoot->incrementChildCount();
                superRoot->addChild(taskQueues[i]->getRootTasks()[j]);
                taskQueues[i]->getRootTasks()[j]->setParent(superRoot);
            }
            rootToUse = superRoot;
        }
        else
        {
	    // Use the only root task
            rootToUse = (SleepTask*)taskQueues[i]->getRootTasks()[0];
        }
	
	// Find the perfect minimum time to calculate the tree
        double minPossibleTheoretical = findMinimumPossibleTime(rootToUse);

// ----------------------------------------------------------------------------------

	// Do other test
        printf("Num root tasks: %d\n", taskQueues[i]->getRootTasks().size());
        //      for(int j = 0; j < taskQueues[i]->getRootTasks().size(); j++)
        //      {
        // Find height of network
        int height = findHeight(rootToUse);

        // Calculate num segments and time for each level in network
        std::pair<int,unsigned int> levelSegments[height];
        std::pair<int,double> levelTime[height];
        unsigned int totSegments = sumSegments(0, levelSegments, rootToUse);
        double Ts = sumTime(0, levelTime, rootToUse);
        double averageSegTime =  (double)totSegments/Ts;
        printf("Total Segments: %d, Total time: %f, Average time pr. segment: %f\n", totSegments, Ts, averageSegTime);

        // Sum serial time
        sumTs += Ts;

        // Get info from calculation
        double Ti = 0.0;
        double Ri = 0.0;
        double sumTiRi = 0.0;
        int totNumPipes = 0;
        int numProcs = g_localNumThreads;
        for(int j = 0; j < height; j++)
        {
            totNumPipes += levelSegments[j].first;

            double avgSegTimeCurrLvl = levelTime[j].second/(double)levelSegments[j].second;
            double maxProcs = (double)(levelSegments[j].first >= numProcs) ? numProcs : levelSegments[j].first;

            // Get average pipe calculation time for current level and pipes/procs ratio
            Ti = ((double)levelSegments[j].second * avgSegTimeCurrLvl) / (double)levelSegments[j].first;
            Ri = (double)levelSegments[j].first/maxProcs;

            // Sum results
            sumTiRi += Ti*Ri;

            printf("Level: %d, Pipes: %d, Segments: %d, Time: %f, Average time pr. segment: %f\n", j,
                   levelSegments[j].first, levelSegments[j].second, levelTime[j].second, avgSegTimeCurrLvl);
            printf("Ti: %f, Ri: %f\n", Ti, Ri);
        }

        // Calculate final formula and print results
        double Tp = minPossibleTheoretical; //sumTiRi;
//          double Tp = sumTiRi;

        // Sum parallel time
        sumTp += Tp;

        printf("\nTotal parallel time: %f sec, Total serial time: %f sec.\nDiff: %f sec. Speedup: %f\n----\n",
               Tp/1000000.0, Ts/1000000.0, (Tp-Ts)/1000000.0, Ts/Tp);
        //      }
    }
    printf("\nTOTAL: CPUs: %d,  SumTs: %f sec, SumTp: %f sec, Parallel Speedup: %f\n\n", g_localNumThreads, sumTs/1000000.0, sumTp/1000000.0, sumTs/sumTp);
}

//------------------------------------------------------------------
// Get the number of CPUs
//------------------------------------------------------------------
int getNumCPUs()
{
    int nProcessorsConfigured = sysconf(_SC_NPROCESSORS_CONF);
    int nProcessorsOnline     = sysconf(_SC_NPROCESSORS_ONLN);

//    if (nProcessorsConfigured != -1)
//    {
//        printf("Processors configured: %ld\n",nProcessorsConfigured);
//    }
//    else
//    {
//        printf("Unable to read configured processor count.\n");
//    }
//
//    if (nProcessorsOnline != -1)
//    {
//        printf("Processors online: %ld\n",nProcessorsOnline);
//    }
//    else
//    {
//        printf("Unable to read online processor count.\n");
//    }
//
    return nProcessorsOnline;
}

//------------------------------------------------------------------
// MAIN
//------------------------------------------------------------------
int main(int argc, char **argv)
{
    g_localNumThreads = getNumCPUs();
    g_numTasks = 20;

    if (1 == atoi(argv[1]))
    {
        g_localNumThreads = atoi(argv[2]);
        string file = argv[3];

        // List of task queues
        std::vector< TaskQueue* > taskQueues;

        // Create a thread handler
        ThreadHandler::Instance()->addThreads(g_localNumThreads);

        // Read task file
        readTasksFile(file, taskQueues);

        // Print time used in traversal
        printCalculationTimes(taskQueues);
    }
    else if (2 == atoi(argv[1]))
    {
        string file = argv[3];
        g_localNumThreads = atoi(argv[2]);
        printf("Using %d threads\n", g_localNumThreads);

        // List of task queues
        std::vector< TaskQueue* > taskQueues;

        // Create a thread handler
        ThreadHandler::Instance()->addThreads(g_localNumThreads);

        // Read task file
        readTasksFile(file, taskQueues);

        // Time parallel execution
        ThreadTimer::Instance()->start_us_timer();

        // Distribute work
        for (int i = 0; i < taskQueues.size(); ++i)
        {
            ThreadHandler::Instance()->distributeWork(taskQueues[i]);
            //            printf("\n");
        }

        // Stop timing
        unsigned long parallel_time = ThreadTimer::Instance()->stop_us_timer();

        printf("Parallel time for %d steps: %d. Average: %d\n", taskQueues.size(), parallel_time, parallel_time / taskQueues.size());

        printf("Deleting task queues\n");

        // Delete all queues
        for (int i = 0; i < taskQueues.size(); ++i)
        {
            if (0 != taskQueues[i])
            {
                delete taskQueues[i];
            }
        }
    }
//    else if(4 == argc)
//    {
//        g_localNumThreads = atoi(argv[2]);
//        g_numTasks = atoi(argv[3]);
//
//        printf("Using %d threads on %d tasks\n", g_localNumThreads, g_numTasks);
//
//        g_counter = g_numTasks;
//
//        ThreadHandler::Instance()->addThreads(g_localNumThreads);
//        TaskQueue taskQueue;
//        buildTaskTree(taskQueue, 0, 0);
//        ThreadHandler::Instance()->distributeWork(&taskQueue);
//    }
//

    printf("Deleting thread handler\n");
    // Delete thread handler and threads
    delete ThreadHandler::Instance();

    printf("Deleting thread timer\n");

    // Delete thread timer
    delete ThreadTimer::Instance();

    return 0;
};
